package edu.berkeley.sbp.misc;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import edu.berkeley.sbp.bind.*;
import java.util.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.io.*;

public class MetaGrammar {
    
    public static boolean harsh = true;

    public static void main(String[] args) throws Exception {
        if (args.length != 2) {
            System.err.println("usage: java " + MetaGrammar.class.getName() + " grammarfile.g com.yourdomain.package.ClassName");
            System.exit(-1);
        }
        //StringBuffer sbs = new StringBuffer();
        //((MetaGrammar)new MetaGrammar().walk(meta)).nt.get("e").toString(sbs);
        //System.err.println(sbs);
        String className   = args[1].substring(args[1].lastIndexOf('.')+1);
        String packageName = args[1].substring(0, args[1].lastIndexOf('.'));
        String fileName    = packageName.replace('.', '/') + "/" + className + ".java";

        BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(fileName)));
        StringBuffer out = new StringBuffer();

        boolean skip = false;
        for(String s = br.readLine(); s != null; s = br.readLine()) {
            if (s.indexOf("DO NOT EDIT STUFF BELOW: IT IS AUTOMATICALLY GENERATED") != -1 && s.indexOf("\"")==-1) skip = true;
            if (s.indexOf("DO NOT EDIT STUFF ABOVE: IT IS AUTOMATICALLY GENERATED") != -1 && s.indexOf("\"")==-1) break;
            if (!skip) out.append(s+"\n");
        }

        out.append("\n        // DO NOT EDIT STUFF BELOW: IT IS AUTOMATICALLY GENERATED\n");

        /*
        ReflectiveMeta m = new ReflectiveMeta();
        Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream(args[0])).expand1();
        MetaGrammar.Meta.MetaGrammarFile mgf = m.new MetaGrammarFile(res);
        MetaGrammar.BuildContext bc = new MetaGrammar.BuildContext(mgf);

        Union meta = mgf.get("s").build(bc);
        Tree t = new CharParser(meta).parse(new FileInputStream(args[0])).expand1();
        */
        Tree t = MetaGrammarTree.meta;
        Union u = MetaGrammar.make(t, "s");

        System.err.println();
        System.err.println("== parsing with parsed grammar =================================================================================");
        t = new CharParser((Union)u).parse(new FileInputStream(args[0])).expand1();
        System.out.println(t.toPrettyString());
        //Forest<String> fs = new CharParser(make()).parse(new FileInputStream(args[0]));
        //System.out.println(fs.expand1());

        //GraphViz gv = new GraphViz();
        //fs.toGraphViz(gv);
        //FileOutputStream fox = new FileOutputStream("out.dot");
        //gv.dump(fox);
        //fox.close();

        t.toJava(out);
        out.append("\n        // DO NOT EDIT STUFF ABOVE: IT IS AUTOMATICALLY GENERATED\n");

        for(String s = br.readLine(); s != null; s = br.readLine()) out.append(s+"\n");
        br.close();

        OutputStream os = new FileOutputStream(fileName);
        PrintWriter p = new PrintWriter(new OutputStreamWriter(os));
        p.println(out.toString());
        p.flush();
        os.close();
    }

    public static class ReflectiveMetaPlain extends ReflectiveMeta {
        public Object repeatTag() { return null; }
        public Sequence tryResolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            return null; }
        public Sequence resolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            return Sequence.rewritingSequence(tag, els, labels, drops);
        }
    }

    public static class ReflectiveMeta /*extends MetaGrammar.Meta*/ {
        private final Class _cl;
        private final Class[] _inner;
        public ReflectiveMeta() {
            this(MG.class);
        }
        public ReflectiveMeta(Class c) {
            this._cl = c;
            this._inner = c.getDeclaredClasses();
        }
        public ReflectiveMeta(Class c, Class[] inner) {
            this._cl = c;
            this._inner = inner;
        }
        private boolean match(Method m, String s) { return match(m.getAnnotation(bind.as.class), null, s); }
        private boolean match(bind.as t, Class c, String s) {
            if (t==null) return false;
            if (t.value().equals(s)) return true;
            if (c != null && t.equals("") && c.getSimpleName().equals(s)) return true;
            return false;
        }
        /*
        private boolean match(nonterminal t, Class c, String s) {
            if (t==null) return false;
            if (t.value().equals(s)) return true;
            if (c != null && t.equals("") && c.getSimpleName().equals(s)) return true;
            return false;
        }
        */
        private boolean match(Class c, String s, String nonTerminalName) {
            if (match((bind.as)c.getAnnotation(bind.as.class), c, s)) return true;
            //if (match((nonterminal)c.getAnnotation(bind.as.class), c, nonTerminalName)) return true;
            return false;
        }
        public boolean match(Constructor con, String s, String nonTerminalName) {
            Class c = con.getDeclaringClass();
            if (match((bind.as)con.getAnnotation(bind.as.class), null, s)) return true;
            //if (match((nonterminal)con.getAnnotation(bind.as.class), c, s)) return true;
            return false;
        }
        public Object repeatTag() {
            return new Tree.ArrayBuildingTreeFunctor<Object>();
        }
        public Sequence tryResolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            Production p = new Production(tag, nonTerminalName, els, labels, drops);
            for(Method m : _cl.getMethods())
                if (new Target(m).isCompatible(p))
                    return new Target(m).makeSequence(p);
            for(Class c : _inner)
                for(Constructor con : c.getConstructors())
                    if (new Target(con).isCompatible(p))
                        return new Target(con).makeSequence(p);
            for(Class c : _inner)
                if (new Target(c).isCompatible(p))
                    return new Target(c).makeSequence(p);
            return null;
        }
        public Sequence resolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            Sequence ret = tryResolveTag(tag, nonTerminalName, els, labels, drops);
            if (ret != null) return ret;
            String message = "could not find a Java method/class/ctor matching tag \""+tag+
                "\", nonterminal \""+nonTerminalName+"\" with " + els.length + " arguments";
            if (harsh) {
                throw new RuntimeException(message);
            } else {
                System.err.println(message);
                return Sequence.rewritingSequence(tag, els, labels, drops);
            }
        }
    }

   
    public static class Production {
        public String tag;
        public String nonTerminal;
        public Object[] labels;
        public boolean[] drops;
        public Element[] elements;
        public int count = 0;
        public Production(String tag, String nonTerminal, Element[] elements, Object[] labels, boolean[] drops) {
            this.tag = tag;
            this.elements = elements;
            this.nonTerminal = nonTerminal;
            this.labels = labels;
            this.drops = drops;
            for(int i=0; i<drops.length; i++)
                if (!drops[i])
                    count++;
        }
    }

    public static class Target {
        public int[] buildSequence(Production p) {
            Annotation[][] annotations = _bindable.getArgAnnotations();
            String[]       names       = _bindable.getArgNames();
            String name = _bindable.getSimpleName();
            int len = annotations.length;
            int ofs = 0;
            bind.arg[] argtags  = new bind.arg[len];
            for(int i=0; i<names.length; i++)
                for(Annotation a : annotations[i+ofs])
                    if (a instanceof bind.arg)
                        argtags[i+ofs] = (bind.arg)a;
            return Target.this.buildSequence(p, names, argtags);
        }
        private Bindable _bindable;

        public Target(Object o) { this(Bindable.create(o)); }
        public Target(Bindable b) { this._bindable = b; }

        public String getName() { return _bindable.getSimpleName(); }
        public bind.as getBindAs() { return (bind.as)_bindable.getAnnotation(bind.as.class); }
        //public nonterminal getNonTerminal() { return (nonterminal)_bindable.getAnnotation(bind.as.class); }
        public String toString() { return _bindable.getSimpleName(); }
        public boolean isRaw() { return _bindable.isAnnotationPresent(bind.raw.class); }

        public boolean isCompatible(Production p) {
            bind.as t = getBindAs();
            if (t != null &&
                (t.value().equals(p.tag) ||
                 (t.value().equals("") && getName().equals(p.tag))))
                return buildSequence(p)!=null;

            bind.as n = getBindAs();
            if (n != null &&
                (n.value().equals(p.nonTerminal) ||
                 (n.value().equals("") && getName().equals(p.nonTerminal))))
                return buildSequence(p)!=null;

            return false;
        }

        public int[] buildSequence(Production p, String[] names, bind.arg[] argtags) {
            int argTagged = 0;
            for(int i=0; i<argtags.length; i++)
                if (argtags[i] != null)
                    argTagged++;

            // FIXME: can be smarter here
            if (names.length==p.count) {
                int[] ret = new int[p.count];
                for(int i=0; i<p.count; i++) ret[i] = i;
                return ret;
            } else if (argTagged==p.count) {
                int[] ret = new int[argtags.length];
                int j = 0;
                for(int i=0; i<argtags.length; i++)
                    ret[i] = argtags[i]==null ? -1 : (j++);
                return ret;
            } else {
                return null;
            }
        }
        public Sequence makeSequence(Production p) {
            return Sequence.rewritingSequence(new TargetReducer(buildSequence(p), _bindable, isRaw()),
                                              p.elements, p.labels, p.drops);
        }

    }

    public static class TargetReducer implements Tree.TreeFunctor<Object,Object>, ToJava {
        private int[] map;
        private Bindable _bindable;
        private boolean _israw;

        public void toJava(StringBuffer sb) {
            sb.append("new MetaGrammar.TargetReducer(new int[] {");
            for(int i=0; i<map.length; i++)
                sb.append((i+"")+(i<map.length-1 ? "," : ""));
            sb.append("}, ");
            _bindable.toJava(sb);
            sb.append(", ");
            sb.append(_israw ? "true" : "false");
            sb.append(")");
        }
        
        public TargetReducer(int[] map, Bindable b, boolean raw) {
            this.map = map;
            this._bindable = b;
            this._israw = raw;
        }
        public String toString() { return "reducer-"+_bindable.toString(); }
        public Object invoke(Iterable<Tree<Object>> t) {
            if (_israw) return _bindable.impose(new Object[] { t });
            ArrayList ret = new ArrayList();
            for(Tree tc : t) {
                if (tc.head() != null && tc.head() instanceof Functor)
                    ret.add(((Tree.TreeFunctor<Object,Object>)tc.head()).invoke(tc.children()));
                else if (tc.numChildren() == 0)
                    ret.add(tc.head());
                else {
                    System.err.println("FIXME: don't know what to do about " + tc);
                    ret.add(null);
                }
            }
            System.err.println("input tree: " + t);
            Object[] o = (Object[])ret.toArray(new Object[0]);
            int max = 0;
            for(int i=0; i<map.length; i++) max = Math.max(map[i], max);
            Object[] o2 = new Object[max+1];
            for(int i=0; i<o.length; i++) o2[map[i]] = o[i];
            return _bindable.impose(o2);
        }
    }


    public static Union cached = null;
    public static Union make() {
        /*
        if (cached != null) return cached;
        try {
            ReflectiveMeta m = new ReflectiveMeta();
            Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream("tests/meta.g")).expand1();
            MetaGrammar.Meta.MetaGrammarFile mgf = m.new MetaGrammarFile(res);
            MetaGrammar.BuildContext bc = new MetaGrammar.BuildContext(mgf);
            Union meta = mgf.get("s").build(bc);
            Tree t = new CharParser(meta).parse(new FileInputStream("tests/meta.g")).expand1();
            return cached = make(t, "s");
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        */
        return make(MetaGrammarTree.meta, "s");
    }
    public static Union make(Tree t, String s) { return make(t, s, new ReflectiveMeta()); }
    public static Union make(Tree t, String s, ReflectiveMeta rm) {
        System.out.println("Head: " + t.head());
        Tree.TreeFunctor<Object,Object> red = (Tree.TreeFunctor<Object,Object>)t.head();
        MG.Grammar g = (MG.Grammar)red.invoke(t.children());
        Context cx = new Context(g,rm);
        Union u = null;
        for(MG.NonTerminal nt : g.nonterminals) {
            System.out.println(nt.name);
            Union el = (Union)cx.get(nt.name);
            StringBuffer st = new StringBuffer();
            el.toString(st);
            System.err.println(st);
            if (nt.name.equals(s)) u = el;
        }
        return u;
    }



    public static class MG {
        public static @bind.as Object grammar(Object[] o) {
            if (o==null) return null;
            if (o[0] != null && o[0] instanceof Grammar) return o[0];
            Grammar g = new Grammar();
            g.nonterminals = (NonTerminal[])Reflection.lub((Object[])o);
            return g;
        }
        public static @bind.as class Grammar {
            public NonTerminal get(String s) {
                for(NonTerminal nt : nonterminals)
                    if (nt.name.equals(s))
                        return nt;
                return null;
            }
            public @bind.arg NonTerminal[] nonterminals;
            public String toString() {
                String ret = "[ ";
                for(NonTerminal nt : nonterminals) ret += nt + ", ";
                return ret + " ]";
            }
        }
        public abstract static class Un extends El {
            public Seq[][] sequences;
            public void build(Context cx, Union u) {
                HashSet<Sequence> bad2 = new HashSet<Sequence>();
                for(int i=0; i<sequences.length; i++) {
                    Seq[] group = sequences[i];
                    Union u2 = new Union();
                    if (sequences.length==1) u2 = u;
                    for(int j=0; j<group.length; j++) {
                        group[j].build(cx, u2, false);
                    }
                    if (sequences.length==1) break;
                    Sequence seq = Sequence.singleton(u2);
                    for(Sequence s : bad2) {
                        s.lame = true;
                        seq = seq.not(s);
                    }
                    u.add(seq);
                    bad2.add(Sequence.singleton(u2));
                }
            }
        }
        public static class NonTerminal extends Un {
            public String  name = null;
            public @bind.as("NonTerminal") NonTerminal(@bind.arg("Word") String name,
                                         @bind.arg("RHS") Seq[][] sequences) {
                this.name = name;
                this.sequences = sequences;
            }
            public Element build(Context cx) { return cx.get(name); }
        }

        public static class AnonUn extends Un {
            public @bind.as("(") AnonUn(Seq[][] sequences) {
                this.sequences = sequences;
            }
            public Element build(Context cx) {
                Union ret = new Union();
                build(cx, ret);
                return ret;
            }
        }

        //public static @bind.as void range(char c) { }
        public static class Range {
            public @bind.as("range") Range(char only) { first = only; last = only; }
            public @bind.as("-")     Range(char first, char last) { this.first = first; this.last = last; }
            public char first;
            public char last;
        }
        public static abstract class El {
            public String getLabel() { return null; }
            public String getOwnerTag() { return null; }
            public boolean drop() { return false; }
            public abstract Element build(Context cx);
        }
        public static class Drop extends El {
            public El e;
            public Drop(El e) { this.e = e; }
            public String getLabel() { return null; }
            public boolean drop() { return true; }
            public String getOwnerTag() { return e.getOwnerTag(); }
            public Element build(Context cx) { return e.build(cx); }
        }
        public static class Label extends El {
            public String label;
            public El e;
            public Label(String label, El e) { this.e = e; this.label = label; }
            public String getLabel() { return label; }
            public String getOwnerTag() { return e.getOwnerTag(); }
            public Element build(Context cx) { return e.build(cx); }
        }
        public static /*abstract*/ class Seq {
            HashSet<Seq> and = new HashSet<Seq>();
            HashSet<Seq> not = new HashSet<Seq>();
            El[] elements;
            El follow;
            String tag = null;
            boolean lame;
            public Seq(El e) { this(new El[] { e }); }
            public Seq(El[] elements) { this.elements = elements; }
            public Seq tag(String tag) { this.tag = tag; return this; }
            public Seq follow(El follow) { this.follow = follow; return this; }
            public Seq dup() {
                Seq ret = new Seq(elements);
                ret.and.addAll(and);
                ret.not.addAll(not);
                ret.follow = follow;
                ret.tag = tag;
                return ret;
            }
            public Seq and(Seq s) { and.add(s); s.lame = true; return this; }
            public Seq andnot(Seq s) { not.add(s); s.lame = true; return this; }
            public Seq separate(El sep) {
                El[] elements = new El[this.elements.length * 2 - 1];
                for(int i=0; i<this.elements.length; i++) {
                    elements[i*2]   = this.elements[i];
                    if (i<this.elements.length-1)
                        elements[i*2+1] = new Drop(sep);
                }
                this.elements = elements;
                return this;
            }
            public Sequence build(Context cx, Union u, boolean lame) {
                Sequence ret = build0(cx, lame || this.lame);
                for(Seq s : and) { Sequence dork = s.build(cx, u, true); ret = ret.and(dork); }
                for(Seq s : not) { Sequence dork = s.build(cx, u, true); ret = ret.not(dork); }
                u.add(ret);
                ret.lame = lame;
                return ret;
            }
            public Sequence build0(Context cx, boolean lame) {
                boolean unwrap = false;
                boolean dropAll = lame;
                if (tag!=null && tag.equals("[]")) unwrap  = true;
                if (tag!=null && "()".equals(tag)) dropAll = true;
                Object[] labels = new Object[elements.length];
                boolean[] drops = new boolean[elements.length];
                Element[] els = new Element[elements.length];
                for(int i=0; i<elements.length; i++) {
                    labels[i] = elements[i].getLabel();
                    drops[i]  = elements[i].drop();
                    els[i] = elements[i].build(cx);
                    if (elements[i].getOwnerTag() != null)
                        tag = elements[i].getOwnerTag();
                }
                Sequence ret = null;
                if (dropAll)     ret = Sequence.drop(els, false);
                else if (unwrap) ret = Sequence.unwrap(els, cx.rm.repeatTag(), drops);
                else if (tag!=null) {
                    ret = cx.rm.resolveTag(tag, cx.cnt, els, labels, drops);
                } else {
                    ret = cx.rm.tryResolveTag(tag, cx.cnt, els, labels, drops);
                    if (ret == null) {
                        int idx = -1;
                        for(int i=0; i<els.length; i++)
                            if (!drops[i])
                                if (idx==-1) idx = i;
                                else throw new Error("multiple non-dropped elements in sequence: " + Sequence.drop(els,false));
                        if (idx != -1) ret = Sequence.singleton(els, idx);
                        else           ret = Sequence.drop(els, false);
                    }
                }
                if (this.follow != null)
                    ret.follow = infer(this.follow.build(cx));
                ret.lame = this.lame;
                return ret;
            }
        }
        public static @bind.as("&")   Seq  and(Seq s,         El[] elements) { return s.and(seq(elements)); }
        public static @bind.as("&~")  Seq  andnot(Seq s,      El[] elements) { return s.andnot(seq(elements)); }
        public static @bind.as("->")  Seq  arrow(Seq s, El e)                { return s.follow(e); }
        public static @bind.as("::")  Seq  tag(String tagname, Seq s)        { return s.tag(tagname); }
        public static @bind.as("/")   Seq  slash(Seq s, El e)                { return s.separate(e); }

        public static @bind.as("ps")  Seq  seq(El[] elements)                { return new Seq(elements); }
        public static @bind.as        Seq  psx(Seq s)                        { return s; }
        public static @bind.as(":")   El   colon(String s, El e)             { return new Label(s, e); }
        public static @bind.as(")")   void close(String foo)                 { throw new Error("not supported"); }
        public static @bind.as("()")  El   epsilon()                         { return new Constant(Union.epsilon); }

        public static @bind.as("nonTerminal") class NonTerminalReference extends El {
            public @bind.arg String nonTerminal;
            public Element build(Context cx) {
                return cx.get(nonTerminal);
            }
        }

        public static class StringLiteral        extends Constant {
            public @bind.as("literal") StringLiteral(String string) { super(CharRange.string(string)); }
            public boolean drop() { return true; }
        }

        public static                     class CharClass            extends El {
            Range[] ranges;
            public @bind.as("[") CharClass(Range[] ranges) { this.ranges = ranges; }
            public Element build(Context cx) {
                edu.berkeley.sbp.util.Range.Set set = new edu.berkeley.sbp.util.Range.Set();
                for(Range r : ranges)
                        set.add(r.first, r.last);
                return CharRange.set(set);
            }
        }

        public static @bind.as("{")           class XTree                 extends El {
            public @bind.arg Seq body;
            public Element build(Context cx) {
                throw new Error();
            }
        }

        public static class Rep extends El {
            public El e, sep;
            public boolean zero, many, max;
            public Rep(El e, El sep, boolean zero, boolean many, boolean max) {
                this.e = e; this.sep = sep; this.zero = zero; this.many = many; this.max = max;}
            public Element build(Context cx) {
                return (!max)
                    ? Sequence.repeat(e.build(cx),        zero, many, sep==null ? null : sep.build(cx), cx.rm.repeatTag())
                    : sep==null
                    ? Sequence.repeatMaximal(infer(e.build(cx)), zero, many,                                   cx.rm.repeatTag())
                    : Sequence.repeatMaximal(e.build(cx),                    zero, many, infer(sep.build(cx)), cx.rm.repeatTag());
            }
        }
        public static class Constant extends El {
            Element constant;
            public Constant(Element constant) { this.constant = constant; }
            public Element build(Context cx) { return constant; }
        }
        public abstract static class PostProcess extends El {
            El e;
            public PostProcess(El e) { this.e = e; }
            public Element build(Context cx) { return postProcess(e.build(cx)); }
            public abstract Element postProcess(Element e);
        }

        // FIXME: it would be nice if we could hoist this into "Rep"
        public static @bind.as("++")  El plusmax(final El e)                     { return new Rep(e, null, false, true, true); }
        public static @bind.as("+")   El plus(final El e)                        { return new Rep(e, null, false, true, false); }
        public static @bind.as("++/") El plusmaxfollow(final El e, final El sep) { return new Rep(e, sep,  false, true, true); }
        public static @bind.as("+/")  El plusfollow(final El e, final El sep)    { return new Rep(e, sep,  false, true, false); }
        public static @bind.as("**")  El starmax(final El e)                     { return new Rep(e, null, true,  true, true); }
        public static @bind.as("*")   El star(final El e)                        { return new Rep(e, null, true,  true, false); }
        public static @bind.as("**/") El starmaxfollow(final El e, final El sep) { return new Rep(e, sep,  true,  true, true); }
        public static @bind.as("*/")  El starfollow(final El e, final El sep)    { return new Rep(e, sep,  true,  true, false); }
        public static @bind.as("?")   El question(final El e)                    { return new Rep(e, null, true,  true, false); }

        public static @bind.as("!")   El bang(final El e)                        { return new Drop(e); }

        public static @bind.as("^")   El caret(final String s) {
            return new Drop(new Constant(CharRange.string(s)) {
                    public String getOwnerTag() { return s; }
                });
        }

        public static @bind.as("~")   El tilde(final El e) {
            return new PostProcess(e) {
                    public Element postProcess(Element e) {
                        return infer((Topology<Character>)Atom.toAtom(e).complement()); 
                    } }; }

        public static @bind.as("^^")  void doublecaret(final El e)                 { throw new Error("not implemented"); }

        //public static @bind.as("(")   El subexpression(Seq[][] rhs)                { return new NonTerminal(rhs); }

        public static @bind.as("Word")    String word(String s) { return s; }
        public static @bind.as("Quoted")  String quoted(String s) { return s; }
        public static @bind.as("escaped") String c(char c) { return c+""; }
        public static @bind.as("\"\"")    String emptystring() { return ""; }
        public static @bind.as("\n")      String retur() { return "\n"; }
        public static @bind.as("\r")      String lf() { return "\r"; }

    }
    public static class Context {
        HashMap<String,Union> map = new HashMap<String,Union>();
        private MG.Grammar grammar;
        public String cnt = null;
        private ReflectiveMeta rm;
        public Context(MG.Grammar g, ReflectiveMeta rm) {
            this.grammar = g;
            this.rm = rm;
        }
        public Union build() {
            Union ret = null;
            for(MG.NonTerminal nt : grammar.nonterminals) {
                Union u = get(nt.name);
                if ("s".equals(nt.name))
                    ret = u;
            }
            return ret;
        }
        public Context(Tree t, ReflectiveMeta rm) {
            this.rm = rm;
            Tree.TreeFunctor<Object,Object> red = (Tree.TreeFunctor<Object,Object>)t.head();
            this.grammar = (MG.Grammar)red.invoke(t.children());
        }
        public Union peek(String name) { return map.get(name); }
        public void  put(String name, Union u) { map.put(name, u); }
        public Union get(String name) {
            Union ret = map.get(name);
            if (ret != null) return ret;
            ret = new Union(name);
            map.put(name, ret);
            MG.NonTerminal nt = grammar.get(name);
            if (nt==null) {
                System.err.println("*** warning could not find " + name);
            } else {
                String old = cnt;
                cnt = name;
                nt.build(this, ret);
                cnt = old;
            }
            return ret;
        }

    }
    /*private*/ static Atom infer(Element e)  { return infer((Topology<Character>)Atom.toAtom(e)); }
    /*private*/ static Atom infer(Topology<Character> t) { return new CharRange(new CharTopology(t)); }
}
